"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const child_process_1 = require("child_process");
const path = require("path");
const fs = require("fs");
const npmRegistryFetch = require("npm-registry-fetch");
const debug_1 = require("./debug");
const logger = require("./logging");
const load_json_file_1 = require("./fs/load-json-file");
const find_package_root_1 = require("./packages/find-package-root");
const debug = debug_1.debug(__filename);
const install = (command) => {
    return new Promise((resolve, reject) => {
        const npmInstall = child_process_1.spawn(command, [], { shell: true, stdio: 'inherit' });
        npmInstall.on('error', (err) => {
            return reject(err);
        });
        npmInstall.on('exit', (code) => {
            if (code !== 0) {
                return reject();
            }
            return resolve(true);
        });
    });
};
exports.installPackages = async (packages) => {
    let isDev = false;
    const currentWorkingDir = process.cwd();
    const isWindows = process.platform === 'win32';
    let command = `npm install ${packages.join(' ')}`;
    if (packages.length === 0) {
        return Promise.resolve(true);
    }
    const hintLocalPath = path.join(currentWorkingDir, 'node_modules', 'hint', 'package.json');
    const global = !fs.existsSync(hintLocalPath);
    if (!global) {
        try {
            const packagePath = find_package_root_1.default(currentWorkingDir);
            const jsonContent = load_json_file_1.default(path.join(packagePath, 'package.json'));
            isDev = jsonContent.devDependencies && jsonContent.devDependencies.hasOwnProperty('hint');
        }
        catch (err) {
            isDev = false;
        }
    }
    command += global ? ' -g' : '';
    command += isDev ? ' --save-dev' : '';
    try {
        debug(`Running command ${command}`);
        logger.log('Installing packages...');
        await install(command);
        return true;
    }
    catch (err) {
        debug(err);
        logger.error(`
There was a problem installing packages.
Please try executing:
    ${!isWindows && global ? 'sudo ' : ''}${command}
            manually to install all the packages.`);
        return false;
    }
};
const filterPackages = (packages, initTerm) => {
    return packages.filter((pkg) => {
        return pkg.name.startsWith(initTerm);
    });
};
const getPackages = (result) => {
    return result.objects.map((obj) => {
        return obj.package;
    });
};
const generateSearchQuery = (searchTerm, from, size = 100) => {
    return `/-/v1/search?text=${searchTerm}&size=${size}${from ? `&from=${from}` : ''}`;
};
exports.search = async (searchTerm) => {
    const result = await npmRegistryFetch.json(generateSearchQuery(searchTerm));
    let total = getPackages(result);
    while (result.total > total.length) {
        const r = await npmRegistryFetch.json(generateSearchQuery(searchTerm, total.length));
        total = total.concat(getPackages(r));
    }
    return total;
};
exports.getOfficialPackages = async (type) => {
    const hints = await exports.search(`@hint/${type}`);
    return filterPackages(hints, `@hint/${type}`);
};
exports.getUnnoficialPackages = async (type) => {
    const hints = await exports.search(`hint-${type}`);
    return filterPackages(hints, `hint-${type}`);
};
